LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;
USE ieee.numeric_std.ALL;

library riscv32;
USE riscv32.constants.ALL;

-- TODO: Output should be 32-bit instruction fed into the inst. decoder

-- Decoder of compressed instructions, as part of RVC (RISC-V Compressed instructions)
-- All RCV instructions expand out to the base ISA, or the F/D extension's instructions
-- This entity provides this expansion
ENTITY compressed_decoder IS
	PORT(
		-- Inputs
		-------------
		instruction : IN STD_LOGIC_VECTOR(15 DOWNTO 0);
		clk : IN STD_LOGIC;

		-- Outputs
		-------------
		reg_dest : OUT STD_LOGIC_VECTOR(ADDR_WIDTHM1 DOWNTO 0);
		reg1 : OUT STD_LOGIC_VECTOR(ADDR_WIDTHM1 DOWNTO 0);
		reg2 : OUT STD_LOGIC_VECTOR(ADDR_WIDTHM1 DOWNTO 0);
	);
END compressed_decoder;

ARCHITECTURE behavioural OF compressed_decoder IS

-- Compressed Instructions sometimes use a register width of 3 bits,
-- Referred to in the official documentation as rs1', rs2', rd'; 
-- Common notation translates this into the word "prime"
CONSTANT REG_PRIME_LEN : integer := 3;
CONSTANT ILLEGAL_INST : STD_LOGIC_VECTOR(15 DOWNTO 0) := (OTHERS => '0');
CONSTANT REG_PRIME_LENM1 : integer := REG_PRIME_LEN - 1;

-- Fields defined by the RVC spec
SIGNAL major_op : STD_LOGIC_VECTOR(1 DOWNTO 0);
SIGNAL r2_prime : STD_LOGIC_VECTOR(REG_PRIME_LENM1 DOWNTO 0);
SIGNAL r1_prime : STD_LOGIC_VECTOR(REG_PRIME_LENM1 DOWNTO 0);
SIGNAL funct6 : STD_LOGIC_VECTOR(5 DOWNTO 0);
SIGNAL funct4 : STD_LOGIC_VECTOR(3 DOWNTO 0);
SIGNAL funct3 : STD_LOGIC_VECTOR(2 DOWNTO 0);
SIGNAL funct2 : STD_LOGIC_VECTOR(1 DOWNTO 0);
SIGNAL reg_prime : STD_LOGIC_VECTOR(REG_PRIME_LENM1 DOWNTO 0);
SIGNAL reg_decomp : STD_LOGIC_VECTOR(ADDR_WIDTHM1 DOWNTO 0);

CONSTANT REG_CONVERT : STD_LOGIC_VECTOR(12 DOWNTO 0) := "000" & x"001";

BEGIN

-- See RISC-V Unprivileged ISA doc, Table 33 for definitions
major_op <= instruction(1 DOWNTO 0);
r2_prime <= instruction(4 DOWNTO 2);
r1_prime <= instruction(9 DOWNTO 7);
funct6 <= instruction(15 DOWNTO 10);
funct4 <= instruction(15 DOWNTO 12);
funct3 <= instruction(15 DOWNTO 13);
funct2 <= instruction(6 DOWNTO 5);


-- Convert compressed register to standard:
-- Prepend the compressed register value with REG_CONVERT
-- This corresponds 000 to Register x8/f8, 001 to x9/f9, etc

compressed_inst : PROCESS(clk) BEGIN
IF(rising_edge(clk)) THEN
	CASE major_op IS
	WHEN "00" =>
		IF(ILLEGAL_INST = instruction) THEN
			--Illegal instruction flag
		END IF;
		CASE funct3 IS =>
		WHEN "000" =>
			--ADDI4SPN (Add Immediate [multiplied by 4] to the x2, save to rd prime)
			addi_flag <= '1';
			reg_dest <= REG_CONVERT & r2_prime;
			reg1 <= x"002";
			addi_uimm <= instruction(10 DOWNTO 7) & 
				     instruction(12 DOWNTO 11) &
				     instruction(5) &
				     instruction(6) & "00";
		WHEN "001" =>
			--FLD
			fld_flag <= '1';
			reg_dest <= REG_CONVERT & r2_prime;
			fld_uimm <= instruction(6) &
				    instruction(5) &
				    instruction(12 DOWNTO 10) & 
				    "000";
			reg1 <= REG_CONVERT & r1_prime;
		WHEN "010" =>
			--LW
			lw_flag <= '1';
			reg_dest <= REG_CONVERT & r2_prime;
			lw_uimm <= instruction(5) &
				   instruction(12 DOWNTO 10) &
				   instruction(6) & "00";
			reg1 <= REG_CONVERT & r1_prime;
		WHEN "011" =>
			--FLW
			flw_flag <= '1';
			reg_dest <= REG_CONVERT & r2_prime;
			flw_uimm <= instruction(5) &
				   instruction(12 DOWNTO 10) &
				   instruction(6) & "00";
		WHEN "100" =>
			--RESERVED
		WHEN "101" =>
			--FSD
			fsd_flag <= '1';
			reg2 <= REG_CONVERT & r2_prime;
			reg1 <= REG_CONVERT & r1_prime;
			fsd_uimm <= instruction(6) &
				    instruction(5) &
				    instruction(12 DOWNTO 10) & 
				    "000";
		WHEN "110" =>
			--SW
			sw_flag <= '1';
			reg2 <= REG_CONVERT & r2_prime;
			reg1 <= REG_CONVERT & r1_prime;
			sw_uimm <= instruction(5) &
				   instruction(12 DOWNTO 10) &
				   instruction(6) & "00";
		WHEN "111" =>
			--FSW
			fsw_flag <= '1';
			reg2 <= REG_CONVERT & r2_prime;
			reg1 <= REG_CONVERT & r1_prime;
			sw_uimm <= instruction(5) &
				   instruction(12 DOWNTO 10) &
				   instruction(6) & "00";
		END CASE;
	WHEN "01" =>
		CASE funct3 IS =>
		WHEN "000" =>
			--NOP/ADDI
			addi_flag <= '1';
			reg_dest <= instruction(11 DOWNTO 7);
			reg1 <= instruction(11 DOWNTO 7);
			addi_imm <= instruction(12) &
				    instruction(6 DOWNTO 2);
		WHEN "001" =>
			--JAL
			jal_flag <= '1';
			jal_imm <= instruction(12) &
			       instruction(8) &
			       instruction(10 DOWNTO 9) &
			       instruction(6) & 
			       instruction(7) &
			       instruction(2)
			       instruction(11) &
			       instruction(5 DOWNTO 3);
		WHEN "010" =>
			--LI
			if("00000" = instruction(11 DOWNTO 7)) THEN
				--HINT
			ELSE
				li_flag <= '1';
				reg_dest <= instruction(11 DOWNTO 7);
				reg1 <= instruction(11 DOWNTO 7);
				li_imm <= instruction(12) &
					  instruction(6 DOWNTO 2);
			END IF;
		WHEN "011" =>
			IF ("00010" = instruction(11 DOWNTO 7)) THEN
				--ADDI16SP
				addi16sp_flag <= '1';
				addi16sp_imm <= instruction(12) &
						instruction(4 DOWNTO 3) &
						instruction(5) &
						instruction(2) &
						instruction(6);
			IF ("00000" = instruction(11 DOWNTO 7)) THEN
				--???
			ELSE
				--LUI
				lui_flag <= '1';
				lui_dest <= instruction(11 DOWNTO 7);
				lui_imm <= instruction(12) &
					   instruction(6 DOWNTO 2) &
					   x"000";
			END IF;
		WHEN "100" =>
			--Ops
			IF('1' = instruction(12)) THEN
				--Reserved instruction, ignore
			ELSE
				reg1 <= REG_CONVERT & r1_prime;
				reg_dest <= REG_CONVERT & r1_prime;
				CASE instruction(11 DOWNTO 10) IS
				WHEN "00" =>
					srli_flag <= '1';
					srli_imm <= instruction(12) &
						    instruciton(6 DOWNTO 2);

				WHEN "01" =>
					srai_flag <= '1';
					srai_imm <= instruction(12) &
						    instruciton(6 DOWNTO 2);
				WHEN "10" =>
					andi_flag <= '1';
					andi_imm <= instruction(12) &
						    instruciton(6 DOWNTO 2);
				WHEN "11" =>
					reg2 <= REG_CONVERT & r2_prime;
					CASE instruction(6 DOWNTO 5) IS
						WHEN "00" =>
							sub_flag <= '1';
						WHEN "01" =>
							xor_flag <= '1';
						WHEN "10" =>
							or_flag <= '1';
						WHEN "11" =>
							and_flag <= '1';
					END CASE;
				END CASE;
			END IF;
		WHEN "101" =>
			--J
			j_flag <= '1';
			j_imm <= instruction(12) &
				 instruction(8) &
				 instruction(10 DOWNTO 9) &
				 instruction(6) &
				 instruction(7) &
				 instruction(2) &
				 instruction(11) &
				 instruction(5 DOWNTO 3);
		WHEN "110" =>
			--BEQZ
			beqz_flag <= '1';
			beqz_imm <= instruction(12) &
				    instruction(6 DOWNTO 5) &
				    instruction(2) &
				    instruction(11 DOWNTO 10) &
				    instruction(4 DOWNTO 3);
			reg1 <= REG_CONVERT & r1_prime;
		WHEN "111" =>
			--BNEZ
			bnez_flag <= '1';
			bnez_imm <= instruction(12) &
				    instruction(6 DOWNTO 5) &
				    instruction(2) &
				    instruction(11 DOWNTO 10) &
				    instruction(4 DOWNTO 3);
			reg1 <= REG_CONVERT & r1_prime;
		END CASE;
	WHEN "10" =>
		CASE funct3 IS =>
		WHEN "000" =>
			IF ('00000' = instruction(11 DOWNTO 7)) THEN
				--HINT
			ElSE
				IF('1' = instruction(12) OR "00000" = instruction(6 DOWNTO 2)) THEN
					--CUSTOM INSTUCTION
				ELSE
					--SLLI
					slli_flag <= '1';
					reg1 <= instruction(11 DOWNTO 7);
					reg_dest <= instruction(11 DOWNTO 7);
					slli_uimm <= '0' & instruction(6 DOWNTO 2);
				END IF;
			END IF;
		WHEN "001" =>
			--FLDSP
			fldsp_flag <= '1';
			reg_dest <= instruction(11 DOWNTO 7);
			fldsp_imm <= instruction(4 DOWNTO 2) &
				     instruction(12) &
				     instruction(6 DOWNTO 5) & "000";
		WHEN "010" =>
			IF ("00000" = instruction(11 DOWNTO 7)) THEN
				--RESERVED
			ELSE
				--LWSP
				lwsp_flag <= '1';
				reg_dest <= instruction(11 DOWNTO 7);
				lwsp_imm <= instruction(3 DOWNTO 2) &
					    instruction(12) &
					    instruction(6 DOWNTO 4) & "00";
			END IF;
		WHEN "011" =>
			--FLWSP
			flwsp_flag <= '1';
			reg_dest <= instruction(11 DOWNTO 7);
			flwsp_imm <= instruction(3 DOWNTO 2) &
				    instruction(12) &
				    instruction(6 DOWNTO 4) & "00";
		WHEN "100" =>
			IF ('1' = instruction(12)) THEN
				IF("00000" = instruction(6 DOWNTO 2)) THEN
					IF("00000" = instruction(11 DOWNTO 7)) THEN
						--EBREAK
						ebreak_flag <= '1';
					ELSE
						--JALR
						jalr_flag <= '1';
						reg1 <= instruction(11 DOWNTO 7);
					END IF;
				ELSE
					--ADD
					reg1 <= instruction(11 DOwNTO 7);
					reg_dest <= instruction(11 DOWNTO 7);
					reg2 <= instruction(6 DOWNTO 2);
				END IF;
			ELSE
				IF("00000" = instruction(11 DOWNTO 7)) THEN
					--Fail
				ELSE
					IF ("00000" = instruction(6 DOWNTO 2)) THEN
						--JR
						jr_flag <= '1';
						reg1 <= instruction(11 DOWNTO 7);
					ELSE
						--MV
						mv_flag <= '1';
						reg_dest <= instruction(11 DOWNTO 7);
						reg2 <= instruction(6 DOWNTO 2);
					END IF;
				END IF;
			END IF;
		WHEN "101" =>
			--FSDSP
			fsdsp_flag <= '1';
			reg2 <= instruction(6 DOWNTO 2);
			fsdsp_imm <= instruction(9 DOWNTO 7) &
				     instruction(12 DOWNTO 10);
		WHEN "110" =>
			--SWSP
			swsp_flag <= '1';
			reg2 <= instruction(6 DOWNTO 2);
			swsp_imm <= instruction(8 DOWNTO 7) &
				    instruction(12 DOWNTO 9);
		WHEN "111" =>
			--FSWSP
			fswsp_flag <= '1';
			reg2 <= instruction(6 DOWNTO 2);
			fswsp_imm <= instruction(8 DOWNTO 7) &
				     instruction(12 DOWNTO 9);
		END CASE;
	WHEN OTHERS =>
		--Invalid
	END CASE;
END IF;
END PROCESS;

END behavioural;
