LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.std_logic_unsigned.ALL;
USE ieee.numeric_std.ALL;
USE ieee.math_real.all;

library riscv32;
USE riscv32.constants.ALL;

-- Register file that reads from 2 addresses and writes to 
-- up to 1 address every clock; register 0 always returns 0
-- Same implementation for both float and int; REG_ADDR_WIDTH 
-- determines the address width, REG_DATA_WIDTH determines data width
-- REG_COUNT; 
--
-- This register file in it's current state is not efficiently compatible with any Bitmanip extensions
ENTITY register_file IS
	GENERIC(
		REG_ADDR_WIDTH  : INTEGER := 5;
		REG_DATA_WIDTH  : INTEGER := XLENM1;
	);
	PORT(
		-- INPUTS
		clk: IN  STD_LOGIC;
		write_enable: IN STD_LOGIC;
		--Read addresses
		register_addr_a: IN STD_LOGIC_VECTOR(REG_ADDR_WIDTH DOWNTO 0);
		register_addr_b: IN STD_LOGIC_VECTOR(REG_ADDR_WIDTH DOWNTO 0);
		--Write address and value to write
		write_val: IN STD_LOGIC_VECTOR(REG_DATA_WIDTH DOWNTO 0);
		register_write_addr: IN STD_LOGIC_VECTOR(REG_ADDR_WIDTH DOWNTO 0);
		-- OUTPUTS
		register_a: OUT STD_LOGIC_VECTOR(REG_DATA_WIDTH DOWNTO 0);
		register_b: OUT STD_LOGIC_VECTOR(REG_DATA_WIDTH DOWNTO 0)
	);
END integer_register_file;

ARCHITECTURE behavioural OF integer_register_file IS

TYPE register_t IS ARRAY(1 TO (2 ** REG_ADDR_WIDTH)) OF STD_LOGIC_VECTOR(REG_DATA_WIDTH DOWNTO 0);
signal registers : register_t;

CONSTANT reg_zero : STD_LOGIC_VECTOR(REG_ADDR_WIDTH DOWNTO 0) := (OTHERS => '0');

BEGIN

register_impl : PROCESS(clk) BEGIN
	IF (rising_edge(clk)) THEN
		IF ('1' = write_enable AND ~(reg_zero = register_write_addr)) THEN
			registers(to_integer(unsigned(register_write_addr))) <= write_val;
		END IF;
		IF(reg_zero = register_addr_a) THEN
			register_a <= reg_zero;
		ELSE
			register_a <= registers(to_integer(unsigned(register_addr_a)));
		END IF;
		IF(reg_zero = register_addr_b) THEN
			register_b <= reg_zero;
		ELSE
			register_b <= registers(to_integer(unsigned(register_addr_b)));
		END IF;
	END IF;
END PROCESS;

END behavioural;
